/*************************************************************************
*
*    File Name:  fifo_sol3.v
*         Date:  05/08/06
*       Author:  John Logue
*        Email:  john.logue@xilinx.com
*      Company:  Xilinx
*
*                Copyright (c) 2005 Xilinx, Inc.
*                All rights reserved
*
*************************************************************************/

// This is an implementation of Peter Alfke's "Different FIFO Fix"
// (Solution number 3).

// Below are the Gray-converted codes for the address MSBs

// ----------------------------------------------------
// | Binary    Gray | Binary+1  Gray | Binary-1  Gray |
// | ------    ---- | --------  ---- | --------  ---- |
// | 0000      0000 | 0001      0001 | 1111      1000 |
// | 0001      0001 | 0010      0011 | 0000      0000 |
// | 0010      0011 | 0011      0010 | 0001      0001 |
// | 0011      0010 | 0100      0110 | 0010      0011 |
// | 0100      0110 | 0101      0111 | 0011      0010 |
// | 0101      0111 | 0110      0101 | 0100      0110 |
// | 0110      0101 | 0111      0100 | 0101      0111 |
// | 0111      0100 | 1000      1100 | 0110      0101 |
// | 1000      1100 | 1001      1101 | 0111      0100 |
// | 1001      1101 | 1010      1111 | 1000      1100 |
// | 1010      1111 | 1011      1110 | 1001      1101 |
// | 1011      1110 | 1100      1010 | 1010      1111 |
// | 1100      1010 | 1101      1011 | 1011      1110 |
// | 1101      1011 | 1110      1001 | 1100      1010 |
// | 1110      1001 | 1111      1000 | 1101      1011 |
// | 1111      1000 | 0000      0000 | 1110      1001 |
// ----------------------------------------------------

`define COUNT_MSB   8          /* set this to the highest bit in WRCOUNT and RDCOUNT that is used */
`define NEXTBITNUM `COUNT_MSB-4

`timescale  100 ps / 10 ps

module fifo_sol3 (ALMOSTEMPTY,
                  ALMOSTFULL,
                  DO,
                  DOP,
                  EMPTY,
                  FULL,
                  RDCOUNT,
                  RDERR,
                  WRCOUNT,
                  WRERR,
                  DI,
                  DIP,
                  RDCLK,
                  RDEN,
                  RST,
                  WRCLK,
                  WREN);

    output        ALMOSTEMPTY;
    output        ALMOSTFULL;
    output [31:0] DO;
    output [3:0]  DOP;
    output        EMPTY;
    output        FULL;
    output [11:0] RDCOUNT;
    output        RDERR;
    output [11:0] WRCOUNT;
    output        WRERR;

    input [31:0]  DI;
    input [3:0]   DIP;
    input         RDCLK;
    input         RDEN;
    input         RST;
    input         WRCLK;
    input         WREN;

    parameter ALMOST_FULL_OFFSET = 12'h080;          // not currently used
    parameter ALMOST_EMPTY_OFFSET = 12'h080;         // not currently used
    parameter DATA_WIDTH = 36;                       // not currently used
    parameter FIRST_WORD_FALL_THROUGH = "FALSE";     // not currently used

    defparam utt.DATA_WIDTH = 36;

    defparam utt.FIRST_WORD_FALL_THROUGH = "FALSE";
    defparam utt.ALMOST_FULL_OFFSET      = 12'h004;    //   4 decimal (sim model does not allow 0)
    defparam utt.ALMOST_EMPTY_OFFSET     = 12'h1FC;    // 508 decimal (sim model does not allow 511)
//  synthesis attribute DATA_WIDTH of utt is "36"
//  synthesis attribute FIRST_WORD_FALL_THROUGH of utt is "FALSE"
//  synthesis attribute ALMOST_FULL_OFFSET of utt is "004"
//  synthesis attribute ALMOST_EMPTY_OFFSET of utt is "1FC"

// ***************** Signal Declarations *****************

wire [3:0]           RDCOUNT4msbs = RDCOUNT[`COUNT_MSB:`COUNT_MSB-3];
wire [3:0]           WRCOUNT4msbs = WRCOUNT[`COUNT_MSB:`COUNT_MSB-3];
wire                 NextBit      = WRCOUNT[`NEXTBITNUM];

wire [3:0]           RAgray;      // Read  Address MSBs     converted to gray and registered
wire [3:0]           WAgray;      // Write Address MSBs     converted to gray and registered
wire [3:0]           WAgray_m1;   // Write Address MSBs - 1 converted to gray and registered
wire [3:0]           WAgray_p1;   // Write Address MSBs + 1 converted to gray and registered
wire [3:0]           WAgray_p2;   // Write Address MSBs + 2 converted to gray and registered
wire [3:0]           WAgray_p3;   // Write Address MSBs + 3 converted to gray and registered

// ***************** Binary to Registered Gray Conversions *****************

bin2gray      ra (.gray_reg(RAgray),    .bin(RDCOUNT4msbs), .clk(RDCLK), .reset(RST));
bin2gray      wa (.gray_reg(WAgray),    .bin(WRCOUNT4msbs), .clk(WRCLK), .reset(RST));
bin2gray_m1 wam1 (.gray_reg(WAgray_m1), .bin(WRCOUNT4msbs), .clk(WRCLK), .reset(RST));
bin2gray_p1 wap1 (.gray_reg(WAgray_p1), .bin(WRCOUNT4msbs), .clk(WRCLK), .reset(RST));
bin2gray_p2 wap2 (.gray_reg(WAgray_p2), .bin(WRCOUNT4msbs), .clk(WRCLK), .reset(RST));
bin2gray_p3 wap3 (.gray_reg(WAgray_p3), .bin(WRCOUNT4msbs), .clk(WRCLK), .reset(RST));

// ***************** ALMOSTFULL Generation *****************

// the equality translations are split into two sections so there is only one LUT
// in the combinatorial paths.

wire af_p1a = (WAgray_p1[1:0] == RAgray[1:0]);
FDC  afp1a (.Q(af_p1a_s), .C(WRCLK), .CLR(RST), .D(af_p1a));

wire af_p1b = (WAgray_p1[3:2] == RAgray[3:2]);
FDC  afp1b (.Q(af_p1b_s), .C(WRCLK), .CLR(RST), .D(af_p1b));

wire af_p1c = af_p1a_s & af_p1b_s;
FDC  aful1 (.Q(AF1), .C(WRCLK), .CLR(RST), .D(af_p1c));


wire af_p2a = (WAgray_p2[1:0] == RAgray[1:0]);
FDC  afp2a (.Q(af_p2a_s), .C(WRCLK), .CLR(RST), .D(af_p2a));

wire af_p2b = (WAgray_p2[3:2] == RAgray[3:2]);
FDC  afp2b (.Q(af_p2b_s), .C(WRCLK), .CLR(RST), .D(af_p2b));

wire af_p2c = af_p2a_s & af_p2b_s;
FDC  aful2 (.Q(AF2), .C(WRCLK), .CLR(RST), .D(af_p2c));


wire af_p3a = (WAgray_p3[1:0] == RAgray[1:0]);
FDC  afp3a (.Q(af_p3a_s), .C(WRCLK), .CLR(RST), .D(af_p3a));

wire af_p3b = (WAgray_p3[3:2] == RAgray[3:2]);
FDC  afp3b (.Q(af_p3b_s), .C(WRCLK), .CLR(RST), .D(af_p3b));

wire af_p3c = af_p3a_s & af_p3b_s;
FDC  aful3 (.Q(AF3), .C(WRCLK), .CLR(RST), .D(af_p3c));


wire af_hold_d = AF1 | AF2 | AF3;
FDC  afhld (.Q(af_hold), .C(WRCLK), .CLR(RST), .D(af_hold_d));

wire af_d = AF1 | AF2 | AF3 | af_hold;
FDC  af (.Q(ALMOSTFULL), .C(WRCLK), .CLR(RST), .D(af_d));

// ***************** FULL Generation *****************

wire f_d = af_d & NextBit;
FDC  ful (.Q(FULL), .C(WRCLK), .CLR(RST), .D(f_d));

// ***************** ALMOSTEMPTY Generation *****************

// the equality translation is split into two sections so there is only one LUT
// in the combinatorial paths.

wire Aempty0 = (WAgray[1:0] == RAgray[1:0]);
FDP  aemp0 (.Q(ALMOSTEMPTY_s0), .C(RDCLK), .PRE(RST), .D(Aempty0));

wire Aempty1 = (WAgray[3:2] == RAgray[3:2]);
FDP  aemp1 (.Q(ALMOSTEMPTY_s1), .C(RDCLK), .PRE(RST), .D(Aempty1));

wire ALMOSTEMPTY_sa = ALMOSTEMPTY_s0 & ALMOSTEMPTY_s1;
FDP  aempa (.Q(ALMOSTEMPTY_a),  .C(RDCLK), .PRE(RST), .D(ALMOSTEMPTY_sa));
FDP  aempap (.Q(ALMOSTEMPTY_ap),  .C(RDCLK), .PRE(RST), .D(ALMOSTEMPTY_a));


wire Aempty2 = (WAgray_m1[1:0] == RAgray[1:0]);
FDP  aemp2 (.Q(ALMOSTEMPTY_s2), .C(RDCLK), .PRE(RST), .D(Aempty2));

wire Aempty3 = (WAgray_m1[3:2] == RAgray[3:2]);
FDP  aemp3 (.Q(ALMOSTEMPTY_s3), .C(RDCLK), .PRE(RST), .D(Aempty3));

wire ALMOSTEMPTY_sb = ALMOSTEMPTY_s2 & ALMOSTEMPTY_s3;
FDP  aempb (.Q(ALMOSTEMPTY_b),  .C(RDCLK), .PRE(RST), .D(ALMOSTEMPTY_sb));
FDP  aempbp (.Q(ALMOSTEMPTY_bp),  .C(RDCLK), .PRE(RST), .D(ALMOSTEMPTY_b));


wire ALMOSTEMPTY_all = ALMOSTEMPTY_a | ALMOSTEMPTY_ap | ALMOSTEMPTY_b | ALMOSTEMPTY_bp;
FDP  aempae (.Q(ALMOSTEMPTY),  .C(RDCLK), .PRE(RST), .D(ALMOSTEMPTY_all));


// ***************** FIFO16 Instantiation *****************


FIFO16 utt (.ALMOSTEMPTY(), // this output not used
            .ALMOSTFULL(),  // this output not used
            .DO(DO),
            .DOP(DOP),
            .EMPTY(EMPTY),
            .FULL(),        // this output not used
            .RDCOUNT(RDCOUNT),
            .RDERR(RDERR),
            .WRCOUNT(WRCOUNT),
            .WRERR(WRERR),
            .DI(DI),
            .DIP(DIP),
            .RDCLK(RDCLK),
            .RDEN(RDEN),
            .RST(RST),
            .WRCLK(WRCLK),
            .WREN(WREN));

endmodule

module bin2gray (gray_reg, bin, clk, reset);

output [3:0] gray_reg;
input  [3:0] bin;
input        clk;
input        reset;

reg    [3:0] gray;  // combinatorial

always @(bin)
begin
  case (bin)
    4'b0000 : gray = 4'b0000;
    4'b0001 : gray = 4'b0001;
    4'b0010 : gray = 4'b0011;
    4'b0011 : gray = 4'b0010;
    4'b0100 : gray = 4'b0110;
    4'b0101 : gray = 4'b0111;
    4'b0110 : gray = 4'b0101;
    4'b0111 : gray = 4'b0100;
    4'b1000 : gray = 4'b1100;
    4'b1001 : gray = 4'b1101;
    4'b1010 : gray = 4'b1111;
    4'b1011 : gray = 4'b1110;
    4'b1100 : gray = 4'b1010;
    4'b1101 : gray = 4'b1011;
    4'b1110 : gray = 4'b1001;
    4'b1111 : gray = 4'b1000;
  endcase
end

FDC r0 (.Q(gray_reg[0]), .C(clk), .CLR(reset), .D(gray[0]));
FDC r1 (.Q(gray_reg[1]), .C(clk), .CLR(reset), .D(gray[1]));
FDC r2 (.Q(gray_reg[2]), .C(clk), .CLR(reset), .D(gray[2]));
FDC r3 (.Q(gray_reg[3]), .C(clk), .CLR(reset), .D(gray[3]));

endmodule

module bin2gray_m1 (gray_reg, bin, clk, reset);

output [3:0] gray_reg;
input  [3:0] bin;
input        clk;
input        reset;

reg    [3:0] gray;  // combinatorial

always @(bin)
begin
  case (bin)
    4'b0000 : gray = 4'b1000;
    4'b0001 : gray = 4'b0000;
    4'b0010 : gray = 4'b0001;
    4'b0011 : gray = 4'b0011;
    4'b0100 : gray = 4'b0010;
    4'b0101 : gray = 4'b0110;
    4'b0110 : gray = 4'b0111;
    4'b0111 : gray = 4'b0101;
    4'b1000 : gray = 4'b0100;
    4'b1001 : gray = 4'b1100;
    4'b1010 : gray = 4'b1101;
    4'b1011 : gray = 4'b1111;
    4'b1100 : gray = 4'b1110;
    4'b1101 : gray = 4'b1010;
    4'b1110 : gray = 4'b1011;
    4'b1111 : gray = 4'b1001;
  endcase
end

FDC r0 (.Q(gray_reg[0]), .C(clk), .CLR(reset), .D(gray[0]));
FDC r1 (.Q(gray_reg[1]), .C(clk), .CLR(reset), .D(gray[1]));
FDC r2 (.Q(gray_reg[2]), .C(clk), .CLR(reset), .D(gray[2]));
FDC r3 (.Q(gray_reg[3]), .C(clk), .CLR(reset), .D(gray[3]));

endmodule

module bin2gray_p1 (gray_reg, bin, clk, reset);

output [3:0] gray_reg;
input  [3:0] bin;
input        clk;
input        reset;

reg    [3:0] gray;  // combinatorial

always @(bin)
begin
  case (bin)
    4'b0000 : gray = 4'b0001;
    4'b0001 : gray = 4'b0011;
    4'b0010 : gray = 4'b0010;
    4'b0011 : gray = 4'b0110;
    4'b0100 : gray = 4'b0111;
    4'b0101 : gray = 4'b0101;
    4'b0110 : gray = 4'b0100;
    4'b0111 : gray = 4'b1100;
    4'b1000 : gray = 4'b1101;
    4'b1001 : gray = 4'b1111;
    4'b1010 : gray = 4'b1110;
    4'b1011 : gray = 4'b1010;
    4'b1100 : gray = 4'b1011;
    4'b1101 : gray = 4'b1001;
    4'b1110 : gray = 4'b1000;
    4'b1111 : gray = 4'b0000;
  endcase
end

// these registers are set to 1's rather than 0's on reset to
// prevent an erroneous ALMOSTFULL status for 1 WRCLK period
// after the trailing edge of RST.
FDP r0 (.Q(gray_reg[0]), .C(clk), .PRE(reset), .D(gray[0]));
FDP r1 (.Q(gray_reg[1]), .C(clk), .PRE(reset), .D(gray[1]));
FDP r2 (.Q(gray_reg[2]), .C(clk), .PRE(reset), .D(gray[2]));
FDP r3 (.Q(gray_reg[3]), .C(clk), .PRE(reset), .D(gray[3]));

endmodule

module bin2gray_p2 (gray_reg, bin, clk, reset);

output [3:0] gray_reg;
input  [3:0] bin;
input        clk;
input        reset;

reg    [3:0] gray;  // combinatorial

always @(bin)
begin
  case (bin)            
    4'b0000 : gray = 4'b0011;
    4'b0001 : gray = 4'b0010;
    4'b0010 : gray = 4'b0110;
    4'b0011 : gray = 4'b0111;
    4'b0100 : gray = 4'b0101;
    4'b0101 : gray = 4'b0100;
    4'b0110 : gray = 4'b1100;
    4'b0111 : gray = 4'b1101;
    4'b1000 : gray = 4'b1111;
    4'b1001 : gray = 4'b1110;
    4'b1010 : gray = 4'b1010;
    4'b1011 : gray = 4'b1011;
    4'b1100 : gray = 4'b1001;
    4'b1101 : gray = 4'b1000;
    4'b1110 : gray = 4'b0000;
    4'b1111 : gray = 4'b0001;
  endcase
end

// these registers are set to 1's rather than 0's on reset to
// prevent an erroneous ALMOSTFULL status for 1 WRCLK period
// after the trailing edge of RST.
FDP r0 (.Q(gray_reg[0]), .C(clk), .PRE(reset), .D(gray[0]));
FDP r1 (.Q(gray_reg[1]), .C(clk), .PRE(reset), .D(gray[1]));
FDP r2 (.Q(gray_reg[2]), .C(clk), .PRE(reset), .D(gray[2]));
FDP r3 (.Q(gray_reg[3]), .C(clk), .PRE(reset), .D(gray[3]));

endmodule

module bin2gray_p3 (gray_reg, bin, clk, reset);

output [3:0] gray_reg;
input  [3:0] bin;
input        clk;
input        reset;

reg    [3:0] gray;  // combinatorial

always @(bin)
begin
  case (bin)            
    4'b0000 : gray = 4'b0010;
    4'b0001 : gray = 4'b0110;
    4'b0010 : gray = 4'b0111;
    4'b0011 : gray = 4'b0101;
    4'b0100 : gray = 4'b0100;
    4'b0101 : gray = 4'b1100;
    4'b0110 : gray = 4'b1101;
    4'b0111 : gray = 4'b1111;
    4'b1000 : gray = 4'b1110;
    4'b1001 : gray = 4'b1010;
    4'b1010 : gray = 4'b1011;
    4'b1011 : gray = 4'b1001;
    4'b1100 : gray = 4'b1000;
    4'b1101 : gray = 4'b0000;
    4'b1110 : gray = 4'b0001;
    4'b1111 : gray = 4'b0011;
  endcase
end

// these registers are set to 1's rather than 0's on reset to
// prevent an erroneous ALMOSTFULL status for 1 WRCLK period
// after the trailing edge of RST.
FDP r0 (.Q(gray_reg[0]), .C(clk), .PRE(reset), .D(gray[0]));
FDP r1 (.Q(gray_reg[1]), .C(clk), .PRE(reset), .D(gray[1]));
FDP r2 (.Q(gray_reg[2]), .C(clk), .PRE(reset), .D(gray[2]));
FDP r3 (.Q(gray_reg[3]), .C(clk), .PRE(reset), .D(gray[3]));

endmodule
