--************************************************************************
-- ************************************************************************
-- *
-- *    File Name:  fifo_third_clk.vhd
-- *         Date:  11/08/06
-- *
-- *       Author:  John Logue
-- *        Email:  john.logue@xilinx.com
-- *      Company:  Xilinx
-- *
-- *                Copyright (c) 2006 Xilinx, Inc.
-- *                All rights reserved
-- *
-- ************************************************************************

-- This module is a FIFO that utilizes the third clock system described 
-- in the answer record

-- This module may be instantiated in place of a FIFO16 module. The only 
-- difference is that this module includes input signal FASTCLK.

LIBRARY IEEE;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

ENTITY fifo_third_clk IS
   GENERIC (
      ALMOST_FULL_OFFSET    :  std_logic_vector(8 DOWNTO 0) := "010000000";    
                       --  these are default values, they may be overridden
      ALMOST_EMPTY_OFFSET   :  std_logic_vector(8 DOWNTO 0) := "010000000"); 
                       --  when this module is instantiated.
   PORT (
      ALMOSTEMPTY             : OUT std_logic;   
      ALMOSTFULL              : OUT std_logic;   
      DO                      : OUT std_logic_vector(31 DOWNTO 0);   
      DOP                     : OUT std_logic_vector(3 DOWNTO 0);   
      EMPTY                   : OUT std_logic;   
      FULL                    : OUT std_logic;   
      RDCOUNT                 : OUT std_logic_vector(11 DOWNTO 0);   
      RDERR                   : OUT std_logic;   
      WRCOUNT                 : OUT std_logic_vector(11 DOWNTO 0);   
      WRERR                   : OUT std_logic;   
      DI                      : IN std_logic_vector(31 DOWNTO 0);   
      DIP                     : IN std_logic_vector(3 DOWNTO 0);   
      RDCLK                   : IN std_logic;   
      RDEN                    : IN std_logic;   
      RST                     : IN std_logic;   --  hold RST active until FASTCLK is stable
      WRCLK                   : IN std_logic;   
      WREN                    : IN std_logic;   
      FASTCLK                 : IN std_logic);   --  400 MHz clock, driven by a BUFG
END fifo_third_clk;

ARCHITECTURE translated OF fifo_third_clk IS

component FIFO16
  generic
  (
    ALMOST_FULL_OFFSET : bit_vector  := X"080";                
    ALMOST_EMPTY_OFFSET : bit_vector := X"080";                
    DATA_WIDTH : integer := 36;
    FIRST_WORD_FALL_THROUGH : boolean := false 
  );
  port
  (
    ALMOSTEMPTY : out std_logic;
    ALMOSTFULL : out std_logic;
    DO : out std_logic_vector (31 downto 0);
    DOP : out std_logic_vector (3 downto 0);
    EMPTY : out std_logic;
    FULL : out std_logic;
    RDCOUNT : out std_logic_vector (11 downto 0);
    RDERR : out std_logic;
    WRCOUNT : out std_logic_vector (11 downto 0);
    WRERR : out std_logic;
    DI : in std_logic_vector (31 downto 0);
    DIP : in std_logic_vector (3 downto 0);
    RDCLK : in std_logic;
    RDEN : in std_logic;
    RST : in std_logic;
    WRCLK : in std_logic;
    WREN : in std_logic
  );
end component;

component fifo_third_clk_flags 
   generic (
      ALMOST_EMPTY_OFFSET     :  std_logic_vector(8 DOWNTO 0) := "010000000"; 
                         --  these are default values, they may be overridden
      ALMOST_FULL_OFFSET      :  std_logic_vector(8 DOWNTO 0) := "010000000";    
      DEPTH      :  std_logic_vector(8 DOWNTO 0) := "111111101";
      RD_WR_MSB      :  std_logic_vector(8 DOWNTO 0) := "000001000");    
   PORT (
      ALMOSTEMPTY             : OUT std_logic;   
      ALMOSTFULL              : OUT std_logic;   
      WREN                    : IN std_logic;   --  FIFO Write Enable
      RDEN                    : IN std_logic;   --  FIFO Read Enable
      RDCLK                   : IN std_logic;   --  FIFO Read clock
      WRCLK                   : IN std_logic;   --  FIFO Write clock
      FASTCLK                 : IN std_logic;   --  fast clock 
      RST                     : IN std_logic);   --  Reset
end component;

BEGIN
   
   -- ************************** Instantiate FIFO16 *************************

   fifo16_inst1 : FIFO16 
       GENERIC MAP(
         ALMOST_FULL_OFFSET => X"00001", -- do not change this line 
         ALMOST_EMPTY_OFFSET => X"001FE", -- set this to 510 (FIFO16 depth - 2) 
         DATA_WIDTH => 36,
         FIRST_WORD_FALL_THROUGH => FALSE 
)

      PORT MAP (
         ALMOSTEMPTY => open,
         ALMOSTFULL => open,
         DO => DO,
         DOP => DOP,
         DI => DI,
         DIP => DIP,
         EMPTY => EMPTY,
         FULL => FULL,
         RDCOUNT => RDCOUNT,
         RDERR => RDERR,
         WRCOUNT => WRCOUNT,
         WRERR => WRERR,
         RDCLK => RDCLK,
         RDEN => RDEN,
         RST => RST,
         WRCLK => WRCLK,
         WREN => WREN);   
   
   -- ************************** Instantiate fifo_third_clk_flags module **************************
   
  ftcf : fifo_third_clk_flags 
      GENERIC MAP (
         ALMOST_EMPTY_OFFSET => ALMOST_EMPTY_OFFSET,
         ALMOST_FULL_OFFSET => ALMOST_FULL_OFFSET)
      PORT MAP (
         ALMOSTEMPTY => ALMOSTEMPTY,
         ALMOSTFULL => ALMOSTFULL,
         WREN => WREN,
         RDEN => RDEN,
         RDCLK => RDCLK,
         WRCLK => WRCLK,
         FASTCLK => FASTCLK,
         RST => RST);   
   
END translated;
