/*************************************************************************
*
*    File Name:  fifo_third_clk_flags.v
*         Date:  11/06/06
*
*       Author:  John Logue
*        Email:  john.logue@xilinx.com
*      Company:  Xilinx
*
*                Copyright (c) 2006 Xilinx, Inc.
*                All rights reserved
*
*************************************************************************/

// If the frequencies of WRCLK and RDCLK are low enough, it is possible to synchronize FIFO reads and writes to a third
// asynchronous fast clock, generate ALMOSTFULL and ALMOSTEMPTY in this fast clock domain, and then resynchronize
// these flags to their respective clocks. 

// This system requires a minimum of 2 and a maximum of 3 fast clock cycles to process a single
// read or write cycle.  To handle back-to-back read or writes, the fast process must complete within 1 RDCLK or WRCLK
//  period.  Thus, the fast clock must be at least 3 times faster than the faster of WRCLK and RDCLK.

// If the fastest RDCLK or WRCLK is 125 MHz, than FASTCLK could be 400 MHz (400/125 = 3.2).

// For more information, see PowerPoint document third_clk_03.ppt 

// This module implements the logic in the PowerPoint document (except for the FASTCLK generator).
// This module is instantiated in module fifo_third_clk.

`timescale  1 ps / 1 ps

`ifdef DEPTH                 // this is usable depth off the FIFO
`else
  `define  DEPTH   509       // not 512 because of "Forbidden Zone"; set to actual FIFO depth - 3
`endif

`ifdef RD_WR_MSB             // this is the most significant used bit number of WRCOUNT & RDCOUNT in FIFO16
`else
  `define RD_WR_MSB  8       // default to FIFO16 with 512 depth
`endif

`define WIF_MSB (`RD_WR_MSB+1)

module fifo_third_clk_flags(ALMOSTEMPTY,
                  ALMOSTFULL,
                  WREN,
                  RDEN,
                  RDCLK,
                  WRCLK,
                  FASTCLK,
                  RST);

parameter ALMOST_EMPTY_OFFSET = 9'h080;   // these are default values, they may be overridden
parameter ALMOST_FULL_OFFSET  = 9'h080;   // when this module is called

output        ALMOSTEMPTY;
output        ALMOSTFULL;
input         WREN;        // FIFO Write Enable
input         RDEN;        // FIFO Read Enable
input         RDCLK;       // FIFO Read clock
input         WRCLK;       // FIFO Write clock
input         FASTCLK;     // fast clock 
input         RST;         // Reset

// --------------------- Write Control ---------------------

reg WR_start_toggle;
reg WR_end_toggle;
reg WM;              // synchronizer
reg Write;

always @ (posedge WRCLK or posedge RST)
begin
  if(RST)       WR_start_toggle <= 0;
  else if(WREN) WR_start_toggle <= ~WR_start_toggle;  
end

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST)        WR_end_toggle <= 0;
  else if(Write) WR_end_toggle <= ~WR_end_toggle;  
end

wire Wr = WR_start_toggle ^ WR_end_toggle;

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) WM <= 0;
  else    WM <= Wr & ~Write;  
end

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) Write <= 0;
  else    Write <= WM & ~Write;  
end

// --------------------- Read Control ---------------------

reg RD_start_toggle;
reg RD_end_toggle;
reg RM;              // synchronizer
reg Read;

always @ (posedge RDCLK or posedge RST)
begin
  if(RST)       RD_start_toggle <= 0;
  else if(RDEN) RD_start_toggle <= ~RD_start_toggle;  
end

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST)       RD_end_toggle <= 0;
  else if(Read) RD_end_toggle <= ~RD_end_toggle;  
end

wire Rd = RD_start_toggle ^ RD_end_toggle;

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) RM <= 0;
  else    RM <= Rd & ~Read;  
end

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) Read <= 0;
  else    Read <= RM & ~Read;  
end

// --------------------- Words in FIFO Accumulator ---------------------

reg [`WIF_MSB:0] WIF;
wire Change_WIF = Write ^ Read;

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST)
  begin
    WIF <= 0;
  end
  else if(Change_WIF)
  begin
    if(Write) WIF <= WIF + 1;
    else      WIF <= WIF - 1;
  end 
end

// --------------------- ALMOSTFULL generator ---------------------

wire af_0;       // comparitor
reg  af_1;       // register comparitor
reg  af_2;       // synchronizer
reg  ALMOSTFULL;

assign af_0 = WIF > (`DEPTH-ALMOST_FULL_OFFSET);

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) af_1 <= 0;
  else    af_1 <= af_0;  
end

always @ (posedge WRCLK or posedge RST)
begin
  if(RST) af_2 <= 0;
  else    af_2 <= af_1;  
end

always @ (posedge WRCLK or posedge RST)
begin
  if(RST) ALMOSTFULL <= 0;
  else    ALMOSTFULL <= af_2;  
end

// --------------------- ALMOSTEMPTY generator ---------------------

wire ae_0;       // comparitor
reg  ae_1;       // register comparitor
reg  ae_2;       // synchronizer
reg  ALMOSTEMPTY;

assign ae_0 = WIF < ALMOST_EMPTY_OFFSET;

always @ (posedge FASTCLK or posedge RST)
begin
  if(RST) ae_1 <= 1;
  else    ae_1 <= ae_0;  
end

always @ (posedge WRCLK or posedge RST)
begin
  if(RST) ae_2 <= 1;
  else    ae_2 <= ae_1;  
end

always @ (posedge WRCLK or posedge RST)
begin
  if(RST) ALMOSTEMPTY <= 1;
  else    ALMOSTEMPTY <= ae_2;  
end

endmodule
