--************************************************************************
-- ************************************************************************
-- *
-- *    File Name:  fifo_third_clk_flags.vhd
-- *         Date:  11/06/06
-- *
-- *       Author:  John Logue
-- *        Email:  john.logue@xilinx.com
-- *      Company:  Xilinx
-- *
-- *                Copyright (c) 2006 Xilinx, Inc.
-- *                All rights reserved
-- *
-- ************************************************************************
-- If the frequencies of WRCLK and RDCLK are low enough, it is possible to 
-- synchronize FIFO reads and writes to a third asynchronous fast clock,
-- generate ALMOSTFULL and ALMOSTEMPTY in this fast clock domain, and then 
-- resynchronize these flags to their respective clocks. 

-- This system requires a minimum of 2 and a maximum of 3 fast clock cycles to 
-- process a single read or write cycle.  To handle back-to-back read or writes, 
-- the fast process must complete within 1 RDCLK or WRCLK period.  Thus, the 
-- fast clock must be at least 3 times faster than the faster of WRCLK and 
-- RDCLK.

-- If the fastest RDCLK or WRCLK is 125 MHz, than FASTCLK could be 400 MHz 
-- (400/125 = 3.2).
-- This module implements the logic in the solution document (except for the 
-- FASTCLK generator).
-- This module is instantiated in module fifo_third_clk.

LIBRARY IEEE;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

ENTITY fifo_third_clk_flags IS
   GENERIC (
      ALMOST_EMPTY_OFFSET     :  std_logic_vector(8 DOWNTO 0) := "010000000"; 
                         --  these are default values, they may be overridden
      ALMOST_FULL_OFFSET      :  std_logic_vector(8 DOWNTO 0) := "010000000";    
      DEPTH      :  std_logic_vector(8 DOWNTO 0) := "111111101";
      RD_WR_MSB      :  std_logic_vector(8 DOWNTO 0) := "000001000");    
   PORT (
      ALMOSTEMPTY             : OUT std_logic;   
      ALMOSTFULL              : OUT std_logic;   
      WREN                    : IN std_logic;   --  FIFO Write Enable
      RDEN                    : IN std_logic;   --  FIFO Read Enable
      RDCLK                   : IN std_logic;   --  FIFO Read clock
      WRCLK                   : IN std_logic;   --  FIFO Write clock
      FASTCLK                 : IN std_logic;   --  fast clock 
      RST                     : IN std_logic);   --  Reset
END fifo_third_clk_flags;

ARCHITECTURE translated OF fifo_third_clk_flags IS


   -- --------------------- Write Control ---------------------
   SIGNAL WR_start_toggle          :  std_logic;   
   SIGNAL WR_end_toggle            :  std_logic;   
   SIGNAL WM                       :  std_logic;   --  synchronizer
   SIGNAL Write                    :  std_logic;   
   SIGNAL Wr                       :  std_logic;   
   -- --------------------- Read Control ---------------------
   SIGNAL RD_start_toggle          :  std_logic;   
   SIGNAL RD_end_toggle            :  std_logic;   
   SIGNAL RM                       :  std_logic;   --  synchronizer
   SIGNAL Read                     :  std_logic;   
   SIGNAL Rd                       :  std_logic;   
   -- --------------------- Words in FIFO Accumulator ---------------------
   SIGNAL WIF                      :  std_logic_vector((8 + 1) DOWNTO 0)
   ;   
   SIGNAL Change_WIF               :  std_logic;   
   -- --------------------- ALMOSTFULL generator ---------------------
   SIGNAL af_0                     :  std_logic;   --  comparitor
   SIGNAL af_1                     :  std_logic;   --  register comparitor
   SIGNAL af_2                     :  std_logic;   --  synchronizer
   -- --------------------- ALMOSTEMPTY generator ---------------------
   SIGNAL ae_0                     :  std_logic;   --  comparitor
   SIGNAL ae_1                     :  std_logic;   --  register comparitor
   SIGNAL ae_2                     :  std_logic;   --  synchronizer

BEGIN

   PROCESS (WRCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         WR_start_toggle <= '0';    
      ELSIF (WRCLK'EVENT AND WRCLK = '1') THEN
         IF (WREN = '1') THEN
            WR_start_toggle <= NOT WR_start_toggle;    
         END IF;
      END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         WR_end_toggle <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         IF (Write = '1') THEN
            WR_end_toggle <= NOT WR_end_toggle;    
         END IF;
      END IF;
   END PROCESS;
   Wr <= WR_start_toggle XOR WR_end_toggle ;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         WM <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         WM <= Wr AND NOT Write;    
      END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         Write <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         Write <= WM AND NOT Write;    
      END IF;
   END PROCESS;

   PROCESS (RDCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         RD_start_toggle <= '0';    
      ELSIF (RDCLK'EVENT AND RDCLK = '1') THEN
         IF (RDEN = '1') THEN
            RD_start_toggle <= NOT RD_start_toggle;    
         END IF;
      END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         RD_end_toggle <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         IF (Read = '1') THEN
            RD_end_toggle <= NOT RD_end_toggle;    
         END IF;
      END IF;
   END PROCESS;
   Rd <= RD_start_toggle XOR RD_end_toggle ;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         RM <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         RM <= Rd AND NOT Read;    
      END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         Read <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         Read <= RM AND NOT Read;    
      END IF;
   END PROCESS;
   Change_WIF <= Write XOR Read ;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         WIF <= "0000000000";    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         IF (Change_WIF = '1') THEN
            IF (Write = '1') THEN
               WIF <= WIF + "0000000001";    
            ELSE
               WIF <= WIF - "0000000001";    
            END IF;
         END IF;
      END IF;
   END PROCESS;
   
   --af_0 <= CONV_STD_LOGIC(WIF > ("111111101" - ALMOST_FULL_OFFSET)) ;

   PROCESS(WIF)
   BEGIN
   IF (WIF > ("111111101" - ALMOST_FULL_OFFSET)) then 
   af_0 <= '1';
   ELSE
   af_0 <= '0';
   END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         af_1 <= '0';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         af_1 <= af_0;    
      END IF;
   END PROCESS;

   PROCESS (WRCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         af_2 <= '0';    
      ELSIF (WRCLK'EVENT AND WRCLK = '1') THEN
         af_2 <= af_1;    
      END IF;
   END PROCESS;

   PROCESS (WRCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         ALMOSTFULL <= '0';    
      ELSIF (WRCLK'EVENT AND WRCLK = '1') THEN
         ALMOSTFULL <= af_2;    
      END IF;
   END PROCESS;
   
--ae_0 <= CONV_STD_LOGIC(WIF < ALMOST_EMPTY_OFFSET) ;

   PROCESS(WIF)
   BEGIN
   IF (WIF < ALMOST_EMPTY_OFFSET) then 
   ae_0 <= '1';
   ELSE
   ae_0 <= '0';
   END IF;
   END PROCESS;

   PROCESS (FASTCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         ae_1 <= '1';    
      ELSIF (FASTCLK'EVENT AND FASTCLK = '1') THEN
         ae_1 <= ae_0;    
      END IF;
   END PROCESS;

   PROCESS (WRCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         ae_2 <= '1';    
      ELSIF (WRCLK'EVENT AND WRCLK = '1') THEN
         ae_2 <= ae_1;    
      END IF;
   END PROCESS;

   PROCESS (WRCLK, RST)
   BEGIN
      IF (RST = '1') THEN
         ALMOSTEMPTY <= '1';    
      ELSIF (WRCLK'EVENT AND WRCLK = '1') THEN
         ALMOSTEMPTY <= ae_2;    
      END IF;
   END PROCESS;

END translated;
